
class TABLE_NAMES():
    MAGNETS = 'magnets'
    BPM = 'bpm'
    MULTI_TURN_BPM = 'bpm_multiturn'
    TWISS = 'twiss'
    MACHINE_PARMS = 'machine_params'


class MAGNET_TABLE_TYPES():
    VCOR = 'PeCorV'
    HCOR = 'PeCorH'
    QUADRUPOLE = 'QUADRUPOLE'
    SIXTRUPOLE = 'SIXTRUPOLE'
